@@ -21,7 +21,7 @@
 
 struct smbconf_global global_conf;
 struct smbconf_parser parser;
-struct smbconf_group *global_group;
+struct smbconf_group *global_group, *ipc_group;
 
 unsigned long long memparse(const char *v)
 {
@@ -93,23 +93,20 @@ static int add_new_group(char *line)
 	while (*end && *end != ']')
 		end = g_utf8_find_next_char(end, NULL);
 
-	name = g_ascii_strdown(begin + 1, end - begin - 1);
+	name = g_strndup(begin + 1, end - begin - 1);
 	if (!name)
 		goto out_free;
 
 	lookup = g_hash_table_lookup(parser.groups, name);
 	if (lookup) {
 		parser.current = lookup;
-		pr_info("SMB conf: multiple group definitions `%s'\n",
-				name);
+		pr_info("Multiple definitions for group `%s'\n", name);
 		g_free(name);
 		return 0;
 	}
 
 	group = g_malloc(sizeof(struct smbconf_group));
-	if (!group)
-		goto out_free;
-
+	group->cb_mode = GROUPS_CALLBACK_NONE;
 	group->name = name;
 	group->kv = g_hash_table_new_full(g_str_hash,
 					  g_str_equal,
@@ -152,7 +149,7 @@ static int add_group_key_value(char *line)
 		return 0;
 
 	if (g_hash_table_lookup(parser.current->kv, key)) {
-		pr_info("SMB conf: multuple key-value [%s] %s\n",
+		pr_info("Multiple key-value definitions [%s] %s\n",
 				parser.current->name, key);
 		return 0;
 	}
@@ -195,14 +192,14 @@ static int __mmap_parse_file(const char *fname, int (*callback)(char *data))
 
 	fd = g_open(fname, O_RDONLY, 0);
 	if (fd == -1) {
-		ret = errno;
-		pr_err("Can't open `%s': %s\n", fname, strerr(ret));
-		return -ret;
+		ret = -errno;
+		pr_debug("Can't open `%s': %m\n", fname);
+		return ret;
 	}
 
 	file = g_mapped_file_new_from_fd(fd, FALSE, &err);
 	if (err) {
-		pr_err("%s: `%s'\n", err->message, fname);
+		pr_err("Can't map `%s' to memory: %s\n", fname, err->message);
 		g_error_free(err);
 		ret = -EINVAL;
 		goto out;
@@ -231,15 +228,7 @@ static int __mmap_parse_file(const char *fname, int (*callback)(char *data))
 			if (!sz)
 				break;
 
-			data = g_malloc(sz + 1);
-			if (!data) {
-				ret = -ENOMEM;
-				goto out;
-			}
-
-			strncpy(data, contents, sz);
-			data[sz] = 0x00;
-
+			data = g_strndup(contents, sz);
 			ret = callback(data);
 			if (ret) {
 				g_free(data);
@@ -260,7 +249,7 @@ out:
 	if (fd) {
 		g_close(fd, &err);
 		if (err) {
-			pr_err("%s: %s\n", err->message, fname);
+			pr_err("Can't close `%s': %s\n", fname, err->message);
 			g_error_free(err);
 		}
 	}
@@ -272,7 +261,8 @@ static int init_smbconf_parser(void)
 	if (parser.groups)
 		return 0;
 
-	parser.groups = g_hash_table_new(g_str_hash, g_str_equal);
+	parser.groups = g_hash_table_new(shm_share_name_hash,
+					 shm_share_name_equal);
 	if (!parser.groups)
 		return -ENOMEM;
 	return 0;
@@ -379,8 +369,7 @@ static int cp_add_global_guest_account(gpointer _v)
 
 	user = usm_lookup_user(_v);
 	if (!user) {
-		pr_err("Fatal error: unable to find `%s' account.\n",
-			(const char *) _v);
+		pr_err("Unable to find user `%s'\n", (char *) _v);
 		return -EINVAL;
 	}
 
@@ -572,6 +561,9 @@ static void global_conf_default(void)
 
 static void global_conf_create(void)
 {
+	if (!global_group || global_group->cb_mode != GROUPS_CALLBACK_INIT)
+		return;
+
 	/*
 	 * This will transfer server options to global_conf, and leave behind
 	 * in the global parser group, the options that must be applied to every
@@ -580,10 +572,25 @@ static void global_conf_create(void)
 	g_hash_table_foreach_remove(global_group->kv, global_group_kv, NULL);
 }
 
+static void append_key_value(gpointer _k, gpointer _v, gpointer user_data)
+{
+	GHashTable *receiver = (GHashTable *)user_data;
+
+	/* Don't override local share options */
+	if (!g_hash_table_lookup(receiver, _k))
+		g_hash_table_insert(receiver, g_strdup(_k), g_strdup(_v));
+}
+
+static void global_conf_update(struct smbconf_group *group)
+{
+	if (!global_group)
+		return;
+
+	g_hash_table_foreach(global_group->kv, append_key_value, group->kv);
+}
+
 static void global_conf_fixup_missing(void)
 {
-	int ret;
-
 	/*
 	 * Set default global parameters which were not specified
 	 * in smb.conf
@@ -601,56 +608,39 @@ static void global_conf_fixup_missing(void)
 		global_conf.work_group =
 			cp_get_group_kv_string(KSMBD_CONF_DEFAULT_WORK_GROUP);
 	if (!global_conf.tcp_port)
-		global_conf.tcp_port = KSMBD_CONF_DEFAULT_TPC_PORT;
+		global_conf.tcp_port = KSMBD_CONF_DEFAULT_TCP_PORT;
 
 	if (global_conf.sessions_cap <= 0)
 		global_conf.sessions_cap = KSMBD_CONF_DEFAULT_SESS_CAP;
 
-	if (global_conf.guest_account)
-		return;
-
-	ret = cp_add_global_guest_account(KSMBD_CONF_DEFAULT_GUEST_ACCOUNT);
-	if (!ret)
-		return;
-	ret = cp_add_global_guest_account(KSMBD_CONF_FALLBACK_GUEST_ACCOUNT);
-	if (ret)
-		pr_err("Fatal error: Cannot set a global guest account %d\n",
-			ret);
+	if (!global_conf.guest_account &&
+	    cp_add_global_guest_account(KSMBD_CONF_DEFAULT_GUEST_ACCOUNT))
+		pr_err("Unable to add guest account\n");
 }
 
-static void append_key_value(gpointer _k, gpointer _v, gpointer user_data)
-{
-	GHashTable *receiver = (GHashTable *) user_data;
-
-	/* Don't override local share options */
-	if (!g_hash_table_lookup(receiver, _k))
-		g_hash_table_insert(receiver, g_strdup(_k), g_strdup(_v));
-}
-
-#define GROUPS_CALLBACK_STARTUP_INIT	0x1
-#define GROUPS_CALLBACK_REINIT		0x2
-
 static void groups_callback(gpointer _k, gpointer _v, gpointer user_data)
 {
 	struct smbconf_group *group = (struct smbconf_group *)_v;
+	unsigned short cb_mode = *(unsigned short *)user_data;
 
-	if (group != global_group) {
-		if (global_group && g_ascii_strncasecmp(_k, "ipc$", 4))
-			g_hash_table_foreach(global_group->kv,
-					     append_key_value,
-					     group->kv);
+	if (group == global_group)
+		return;
 
-		shm_add_new_share(group);
-	}
+	group->cb_mode = cb_mode;
+
+	if (group != ipc_group)
+		global_conf_update(group);
+
+	shm_add_new_share(group);
 }
 
-static int cp_add_ipc_share(void)
+static int cp_add_ipc_group(void)
 {
 	char *comment = NULL, *guest = NULL;
 	int ret = 0;
 
-	if (g_hash_table_lookup(parser.groups, "ipc$"))
-		return 0;
+	if (ipc_group)
+		return ret;
 
 	comment = g_strdup("comment = IPC share");
 	guest = g_strdup("guest ok = yes");
@@ -662,15 +652,16 @@ static int cp_add_ipc_share(void)
 		ret = -EINVAL;
 		goto out;
 	}
-	return ret;
 
+	ipc_group = g_hash_table_lookup(parser.groups, "ipc$");
 out:
 	g_free(comment);
 	g_free(guest);
 	return ret;
 }
 
-static int __cp_parse_smbconfig(const char *smbconf, GHFunc cb, long flags)
+static int __cp_parse_smbconfig(const char *smbconf, GHFunc cb,
+				unsigned short cb_mode)
 {
 	int ret;
 
@@ -680,35 +671,32 @@ static int __cp_parse_smbconfig(const char *smbconf, GHFunc cb, long flags)
 	if (ret)
 		return ret;
 
-	ret = cp_add_ipc_share();
-	if (!ret) {
-		global_group = g_hash_table_lookup(parser.groups, "global");
+	ret = cp_add_ipc_group();
+	if (ret)
+		goto out;
 
-		if (global_group && (flags == GROUPS_CALLBACK_STARTUP_INIT))
-			global_conf_create();
+	global_group = g_hash_table_lookup(parser.groups, "global");
+	if (global_group)
+		global_group->cb_mode = cb_mode;
 
-		g_hash_table_foreach(parser.groups,
-				     groups_callback,
-				     NULL);
-
-		global_conf_fixup_missing();
-	}
+	global_conf_create();
+	g_hash_table_foreach(parser.groups, groups_callback, &cb_mode);
+	global_conf_fixup_missing();
+out:
 	cp_smbconfig_destroy();
 	return ret;
 }
 
 int cp_parse_reload_smbconf(const char *smbconf)
 {
-	return __cp_parse_smbconfig(smbconf,
-				    groups_callback,
+	return __cp_parse_smbconfig(smbconf, groups_callback,
 				    GROUPS_CALLBACK_REINIT);
 }
 
 int cp_parse_smbconf(const char *smbconf)
 {
-	return __cp_parse_smbconfig(smbconf,
-				    groups_callback,
-				    GROUPS_CALLBACK_STARTUP_INIT);
+	return __cp_parse_smbconfig(smbconf, groups_callback,
+				    GROUPS_CALLBACK_INIT);
 }
 
 int cp_parse_pwddb(const char *pwddb)
@@ -725,9 +713,9 @@ int cp_smbconfig_hash_create(const char *smbconf)
 	return __mmap_parse_file(smbconf, process_smbconf_entry);
 }
 
-int cp_parse_subauth(const char *subauth_path)
+int cp_parse_subauth(void)
 {
-	return __mmap_parse_file(subauth_path, usm_add_subauth_global_conf);
+	return __mmap_parse_file(PATH_SUBAUTH, usm_add_subauth_global_conf);
 }
 
 void cp_smbconfig_destroy(void)
