@@ -101,7 +101,7 @@ static void kill_ksmbd_share(struct ksmbd_share *share)
 {
 	int i;
 
-	pr_debug("Kill share %s\n", share->name);
+	pr_debug("Kill share `%s'\n", share->name);
 
 	for (i = 0; i < KSMBD_SHARE_USERS_MAX; i++)
 		free_user_map(share->maps[i]);
@@ -117,7 +117,7 @@ static void kill_ksmbd_share(struct ksmbd_share *share)
 	free(share->veto_list);
 	free(share->guest_account);
 	g_rw_lock_clear(&share->update_lock);
-	free(share);
+	g_free(share);
 }
 
 static int __shm_remove_share(struct ksmbd_share *share)
@@ -188,7 +188,7 @@ static struct ksmbd_share *new_ksmbd_share(void)
 	struct ksmbd_share *share;
 	int i;
 
-	share = calloc(1, sizeof(struct ksmbd_share));
+	share = g_try_malloc0(sizeof(struct ksmbd_share));
 	if (!share)
 		return NULL;
 
@@ -227,9 +227,50 @@ void shm_destroy(void)
 	g_rw_lock_clear(&shares_table_lock);
 }
 
+static char *shm_casefold_share_name(const char *name, size_t len)
+{
+	char *nfdi_name, *nfdicf_name;
+
+	nfdi_name = g_utf8_normalize(name, len, G_NORMALIZE_NFD);
+	if (!nfdi_name)
+		goto out_ascii;
+
+	nfdicf_name = g_utf8_casefold(nfdi_name, strlen(nfdi_name));
+	g_free(nfdi_name);
+	return nfdicf_name;
+out_ascii:
+	g_free(nfdi_name);
+	return g_ascii_strdown(name, len);
+}
+
+guint shm_share_name_hash(gconstpointer name)
+{
+	char *cf_name;
+	guint hash;
+
+	cf_name = shm_casefold_share_name(name, strlen(name));
+	hash = g_str_hash(cf_name);
+	g_free(cf_name);
+	return hash;
+}
+
+gboolean shm_share_name_equal(gconstpointer lname, gconstpointer rname)
+{
+	char *cf_lname, *cf_rname;
+	gboolean equal;
+
+	cf_lname = shm_casefold_share_name(lname, strlen(lname));
+	cf_rname = shm_casefold_share_name(rname, strlen(rname));
+	equal = g_str_equal(cf_lname, cf_rname);
+	g_free(cf_lname);
+	g_free(cf_rname);
+	return equal;
+}
+
 int shm_init(void)
 {
-	shares_table = g_hash_table_new(g_str_hash, g_str_equal);
+	shares_table = g_hash_table_new(shm_share_name_hash,
+					shm_share_name_equal);
 	if (!shares_table)
 		return -ENOMEM;
 	g_rw_lock_init(&shares_table_lock);
@@ -256,7 +297,7 @@ struct ksmbd_share *shm_lookup_share(char *name)
 	return share;
 }
 
-static GHashTable *parse_list(GHashTable *map, char **list)
+static GHashTable *parse_list(GHashTable *map, char **list, char grc)
 {
 	int i;
 
@@ -276,6 +317,15 @@ static GHashTable *parse_list(GHashTable *map, char **list)
 		if (!p)
 			continue;
 
+		if (*p == grc) {
+			struct group *gr;
+
+			gr = getgrnam(p + 1);
+			if (gr)
+				parse_list(map, gr->gr_mem, 0x00);
+			continue;
+		}
+
 		user = usm_lookup_user(p);
 		if (!user) {
 			pr_info("Drop non-existing user `%s'\n", p);
@@ -283,14 +333,14 @@ static GHashTable *parse_list(GHashTable *map, char **list)
 		}
 
 		if (g_hash_table_lookup(map, user->name)) {
-			pr_debug("User already exists in a map: %s\n", p);
+			pr_debug("User `%s' already exists in a map\n",
+				 user->name);
 			continue;
 		}
 
 		g_hash_table_insert(map, user->name, user);
 	}
 
-	cp_group_kv_list_free(list);
 	return map;
 }
 
@@ -312,9 +362,9 @@ static void force_group(struct ksmbd_share *share, char *name)
 	if (grp) {
 		share->force_gid = grp->gr_gid;
 		if (share->force_gid == KSMBD_SHARE_INVALID_GID)
-			pr_err("Invalid force gid: %u\n", share->force_gid);
+			pr_err("Invalid force GID: %u\n", share->force_gid);
 	} else
-		pr_err("Unable to lookup up /etc/group entry: %s\n", name);
+		pr_err("Unable to lookup up `/etc/group' entry: %s\n", name);
 }
 
 static void force_user(struct ksmbd_share *share, char *name)
@@ -332,10 +382,10 @@ static void force_user(struct ksmbd_share *share, char *name)
 			share->force_gid = passwd->pw_gid;
 		if (share->force_uid == KSMBD_SHARE_INVALID_UID ||
 				share->force_gid == KSMBD_SHARE_INVALID_GID)
-			pr_err("Invalid force uid / gid: %u / %u\n",
+			pr_err("Invalid force UID/GID: %u/%u\n",
 					share->force_uid, share->force_gid);
 	} else {
-		pr_err("Unable to lookup up /etc/passwd entry: %s\n", name);
+		pr_err("Unable to lookup up `/etc/passwd' entry: %s\n", name);
 	}
 }
 
@@ -471,63 +521,91 @@ static void process_group_kv(gpointer _k, gpointer _v, gpointer user_data)
 	}
 
 	if (shm_share_config(k, KSMBD_SHARE_CONF_VALID_USERS)) {
+		char **users_list;
+
+		users_list = cp_get_group_kv_list(v);
 		share->maps[KSMBD_SHARE_VALID_USERS_MAP] =
 			parse_list(share->maps[KSMBD_SHARE_VALID_USERS_MAP],
-				cp_get_group_kv_list(v));
+				   users_list, '@');
 		if (share->maps[KSMBD_SHARE_VALID_USERS_MAP] == NULL)
 			set_share_flag(share, KSMBD_SHARE_FLAG_INVALID);
+		cp_group_kv_list_free(users_list);
 		return;
 	}
 
 	if (shm_share_config(k, KSMBD_SHARE_CONF_INVALID_USERS)) {
+		char **users_list;
+
+		users_list = cp_get_group_kv_list(v);
 		share->maps[KSMBD_SHARE_INVALID_USERS_MAP] =
 			parse_list(share->maps[KSMBD_SHARE_INVALID_USERS_MAP],
-				cp_get_group_kv_list(v));
+				   users_list, '@');
 		if (share->maps[KSMBD_SHARE_INVALID_USERS_MAP] == NULL)
 			set_share_flag(share, KSMBD_SHARE_FLAG_INVALID);
+		cp_group_kv_list_free(users_list);
 		return;
 	}
 
 	if (shm_share_config(k, KSMBD_SHARE_CONF_READ_LIST)) {
+		char **users_list;
+
+		users_list = cp_get_group_kv_list(v);
 		share->maps[KSMBD_SHARE_READ_LIST_MAP] =
 			parse_list(share->maps[KSMBD_SHARE_READ_LIST_MAP],
-				   cp_get_group_kv_list(v));
+				   users_list, '@');
 		if (share->maps[KSMBD_SHARE_READ_LIST_MAP] == NULL)
 			set_share_flag(share, KSMBD_SHARE_FLAG_INVALID);
+		cp_group_kv_list_free(users_list);
 		return;
 	}
 
 	if (shm_share_config(k, KSMBD_SHARE_CONF_WRITE_LIST)) {
+		char **users_list;
+
+		users_list = cp_get_group_kv_list(v);
 		share->maps[KSMBD_SHARE_WRITE_LIST_MAP] =
 			parse_list(share->maps[KSMBD_SHARE_WRITE_LIST_MAP],
-				   cp_get_group_kv_list(v));
+				   users_list, '@');
 		if (share->maps[KSMBD_SHARE_WRITE_LIST_MAP] == NULL)
 			set_share_flag(share, KSMBD_SHARE_FLAG_INVALID);
+		cp_group_kv_list_free(users_list);
 		return;
 	}
 
 	if (shm_share_config(k, KSMBD_SHARE_CONF_ADMIN_USERS)) {
+		char **users_list;
+
+		users_list = cp_get_group_kv_list(v);
 		share->maps[KSMBD_SHARE_ADMIN_USERS_MAP] =
 			parse_list(share->maps[KSMBD_SHARE_ADMIN_USERS_MAP],
-				   cp_get_group_kv_list(v));
+				   users_list, '@');
 		if (share->maps[KSMBD_SHARE_ADMIN_USERS_MAP] == NULL)
 			set_share_flag(share, KSMBD_SHARE_FLAG_INVALID);
+		cp_group_kv_list_free(users_list);
 		return;
 	}
 
 	if (shm_share_config(k, KSMBD_SHARE_CONF_HOSTS_ALLOW)) {
+		char **hosts_list;
+
+		hosts_list = cp_get_group_kv_list(v);
 		share->hosts_allow_map = parse_list(share->hosts_allow_map,
-						    cp_get_group_kv_list(v));
+						    hosts_list, 0x00);
 		if (share->hosts_allow_map == NULL)
 			set_share_flag(share, KSMBD_SHARE_FLAG_INVALID);
+		cp_group_kv_list_free(hosts_list);
 		return;
 	}
 
 	if (shm_share_config(k, KSMBD_SHARE_CONF_HOSTS_DENY)) {
+		char **hosts_list;
+
+		hosts_list = cp_get_group_kv_list(v);
 		share->hosts_deny_map = parse_list(share->hosts_deny_map,
-						   cp_get_group_kv_list(v));
+						   hosts_list, 0x00);
 		if (share->hosts_deny_map == NULL)
 			set_share_flag(share, KSMBD_SHARE_FLAG_INVALID);
+		cp_group_kv_list_free(hosts_list);
 		return;
 	}
 
@@ -602,9 +680,12 @@ static void init_share_from_group(struct ksmbd_share *share,
 	set_share_flag(share, KSMBD_SHARE_FLAG_OPLOCKS);
 	set_share_flag(share, KSMBD_SHARE_FLAG_STORE_DOS_ATTRS);
 
-	if (!cp_key_cmp(share->name, "IPC$"))
+	if (!g_ascii_strcasecmp(share->name, "ipc$"))
 		set_share_flag(share, KSMBD_SHARE_FLAG_PIPE);
 
+	if (group->cb_mode == GROUPS_CALLBACK_REINIT)
+		set_share_flag(share, KSMBD_SHARE_FLAG_UPDATE);
+
 	g_hash_table_foreach(group->kv, process_group_kv, share);
 
 	fixup_missing_fields(share);
@@ -620,7 +701,7 @@ int shm_add_new_share(struct smbconf_group *group)
 
 	init_share_from_group(share, group);
 	if (test_share_flag(share, KSMBD_SHARE_FLAG_INVALID)) {
-		pr_err("Invalid share %s\n", share->name);
+		pr_err("Share `%s' is invalid\n", share->name);
 		kill_ksmbd_share(share);
 		return 0;
 	}
@@ -628,11 +709,12 @@ int shm_add_new_share(struct smbconf_group *group)
 	g_rw_lock_writer_lock(&shares_table_lock);
 	if (__shm_lookup_share(share->name)) {
 		g_rw_lock_writer_unlock(&shares_table_lock);
-		pr_info("share exists %s\n", share->name);
+		pr_info("Share `%s' already exists\n", share->name);
 		kill_ksmbd_share(share);
 		return 0;
 	}
 
+	pr_debug("New share `%s'\n", share->name);
 	if (!g_hash_table_insert(shares_table, share->name, share)) {
 		kill_ksmbd_share(share);
 		ret = -EINVAL;
@@ -671,7 +753,7 @@ int shm_lookup_hosts_map(struct ksmbd_share *share,
 			  enum share_hosts map,
 			  char *host)
 {
-	GHashTable *lookup_map;
+	GHashTable *lookup_map = NULL;
 	int ret = -ENOENT;
 
 	if (map >= KSMBD_SHARE_HOSTS_MAX) {
@@ -758,6 +840,8 @@ int shm_handle_share_config_request(struct ksmbd_share *share,
 	resp->force_directory_mode = share->force_directory_mode;
 	resp->force_uid = share->force_uid;
 	resp->force_gid = share->force_gid;
+	*resp->share_name = 0x00;
+	strncat(resp->share_name, share->name, KSMBD_REQ_MAX_SHARE_NAME - 1);
 	resp->veto_list_sz = share->veto_list_sz;
 
 	if (test_share_flag(share, KSMBD_SHARE_FLAG_PIPE))
